//analyze_condition
#include "die.h"
#include "analyze_condition.h"
#include "ev.h"
#include <stdlib.h>
#include <string.h>
#include "analyze_output.h"
#include "mtanalyze.h"

class ConditionWaiters {
	struct Waiter {
		Waiter *next;
		Object *thr;
		const EventRecord *wait_event;
	};
	Waiter *first_waiter;
	size_t waiter_count;
public:
	ConditionWaiters() : first_waiter(0), waiter_count(0) {}
	~ConditionWaiters() {
		Waiter *w=first_waiter;
		while(w) {
			Waiter *n=w->next;
			delete w;
			w=n;
		}
	}
	
	void add(Object *thr, const EventRecord *wait_event);
	size_t waiters() const { return waiter_count; }
	void remove(Object *thr);
};

void ConditionWaiters::add(Object *thr, const EventRecord *wait_event) {
	Waiter *w=new Waiter;
	w->next = first_waiter;
	w->thr = thr;
	w->wait_event = wait_event;
	first_waiter = w;
	waiter_count++;
}

void ConditionWaiters::remove(Object *thr) {
	waiter_count--;
	for(Waiter *p=first_waiter,*pp=0; p; pp=p,p=p->next) {
		if(p->thr==thr) {
			if(pp)
				pp->next = p->next;
			else
				first_waiter = p->next;
			delete p;
			return;
		}
	}
}


struct Condition {
	Object *usual_mutex;
	Object *current_mutex;
	unsigned accum_wait_operations;
	unsigned accum_timedwait_operations;
	unsigned accum_failed_timedwait_operations;
	unsigned accum_signal_operations;
	unsigned accum_broadcast_operations;
	ConditionWaiters cw;
	unsigned accum_signal_ops_without_waiters;
	unsigned accum_broadcast_ops_with_multiple_waiters;
	unsigned long accum_waiters_when_broadcast;
	bool has_warned_about_different_mutexes;
	bool has_warned_about_signal_without_mutex;
};


static void analyze1(const EventRecord ev[], size_t events,
                     Object *o_cond, size_t conds,
		     Condition *cond,
		     EventObjectPointers *eop)
{
	for(size_t i=0; i<events; i++) {
		const EventRecord *ep = ev+i;
		Object *o=eop[i].cond;
		if(!o) continue;
		Condition *c = cond+(o-o_cond);
		switch(ep->event) {
			case ev_cond_broadcast: {
				c->accum_broadcast_operations++;
				size_t waiters = c->cw.waiters();
				if(waiters>1)
					c->accum_broadcast_ops_with_multiple_waiters++;
				c->accum_waiters_when_broadcast += waiters;
				break;
			}
			case ev_cond_broadcasted: {
				//check rc
				if(ep->info!=0)
					warn(ep,CONDBROADCASTERROR,"pthread_cond_broadcast() returned an error %d. This is unusual",ep->info);
				break;
			}
			case ev_cond_destroy: {
				//check rc
				if(ep->info!=0)
					warn(ep,CONDDESTROYERROR,"pthread_cond_destroy() returned an error %d. This is unusual",ep->info);
				break;
			}
			case ev_cond_destroyed: {
				if(c->cw.waiters()>0)
					warn(ep,CONDDESTROYINUSE,"Condition destroyed while %ld waiters exist",(long)c->cw.waiters());
				break;
			}
			case ev_cond_create: {
				//nop
				break;
			}
			case ev_cond_created: {
				//check rc
				break;
			}
			case ev_cond_create_failed: {
				warn(ep,CONDINITERROR,"pthread_cond_init() returned an error %d. This is unusual",ep->info);
				break;
			}
			case ev_cond_signal: {
				c->accum_signal_operations++;
				size_t waiters = c->cw.waiters();
				if(waiters==0)
					c->accum_signal_ops_without_waiters++;
				else if(!c->has_warned_about_signal_without_mutex) {
					//check mutex ownership
					unsigned long mtx_addr = c->current_mutex->addr;
					for(size_t j=i; j--!=0; ) {
						if(eop[j].thr==eop[i].thr &&
						   ev[j].mtx_addr==mtx_addr)
						{
							if(ev[j].event==ev_mutex_locked)
								break; //OK
							warn(ep,CONDSIGNALWITHOUTMUTEXLOCK,
							     "(this warning will appear only once for this condition)\n"
							     "pthread_cond_signal() called without holding the mutex (addr=%08lx) which is used in pthread_cond_wait/timedwait() for this condition\n",
							     mtx_addr);
							c->has_warned_about_signal_without_mutex = true;
							break;
						}
					}
				}
				break;
			}
			case ev_cond_signalled: {
				//check rc
				if(ep->info!=0)
					warn(ep,CONDSIGNALERROR,"pthread_cond_signal() returned an error. This is unusual.\n");
				break;
			}
			case ev_cond_wait:
			case ev_cond_timedwait: {
				//check mutex
				//check usual
				//set current
				//register waiter
				if(c->cw.waiters()==0) {
					if(c->usual_mutex &&
					   c->usual_mutex!=eop[i].mtx &&
					   !c->has_warned_about_different_mutexes)
					{
						warn(ep,CONDWAITNOTUSUALMUTEX,
						     "(this warning will appear only once for this condition)\n"
						     "pthread_cond_wait() called with a different mutex than usual (usual mutex = %08lx). This is probably not intended.\n",
						     eop[i].mtx->addr
						    );
						c->has_warned_about_different_mutexes=true;
					}
					c->usual_mutex = eop[i].mtx;
					c->current_mutex = eop[i].mtx;
				} else {
					if(c->current_mutex!=eop[i].mtx)
						warn(ep,CONDWAITWRONGMUTEX,
						     "pthread_cond_wait/timedwait() is called with mutex different from specified in other current wait/timedwait operations (mutex addr=%08lx)\n"
						     "This leads to undefined behaviour.\n",
						     c->current_mutex->addr
						    );
				}
				
				c->accum_wait_operations++;
				if(ep->event==ev_cond_wait)
					c->accum_timedwait_operations++;
				
				c->cw.add(eop[i].thr,ep);
				break;
			}
			
			case ev_cond_waited: {
				//check rc (unusual)
				//deregister waiter
				if(ep->info!=0)
					warn(ep,CONDWAITERROR,
					     "pthread_cond_wait() returned the error %d. This is unusual\n",
					     ep->info
					    );
				c->cw.remove(eop[i].thr);
				break;
			}
			case ev_cond_timedwait_failed: {
				//increment failed timedwait
				//deregister waiter
				c->accum_failed_timedwait_operations++;
				c->cw.remove(eop[i].thr);
				break;
			}
			case ev_cond_name:
				break;
			default:
				die(ep,COND_UNKNOWN_OP,
				    "(internal error) condition address given, but event type is not handled\n");
		} //switch
	}
	//check for waiters
	for(size_t i=0; i<conds; i++) {
		if(cond[i].cw.waiters()>0)
			suggest(0,CONDSTILLWAITING,
			        "Condition %08lx has %ld waiting threads at the end of the event log\n",
				o_cond[i].addr,
				cond[i].cw.waiters()
			       );	
	}
}


static void look_for_thundering_herd(const EventRecord ev[], size_t events,
                                     Object *o_cond,
                                     Condition *cond,
                                     EventObjectPointers eop[]
                                    )
{
	size_t e_start,e_end;
	if(o_cond->create_event)
		e_start = (size_t)(o_cond->create_event-ev);
	else
		e_start = 0;
	if(o_cond->destroy_event)
		e_end = (size_t)(o_cond->destroy_event-ev);
	else
		e_end = events;
	
	const EventRecord *ev_thundering=0;
	
	unsigned long short_wakeups=0;
	for(size_t i=e_start; i<e_end; i++) {
		if(eop[i].cond!=o_cond ||
		   ev[i].event!=ev_cond_broadcast) continue;
		//find wakeups
		for(size_t j=i+1; j<e_end; j++) {
			if(eop[j].cond!=o_cond) continue;
			if(ev[j].event==ev_cond_signal ||
			   ev[j].event==ev_cond_broadcast) break;
			if(ev[j].event==ev_cond_waited) {
				//found a wakeup
				//does the thread wait again on this condition without doing much?
				double endtime=ev[j].timestamp+warn_thundering_herd_max_intervening_time;
				unsigned ops=0;
				size_t k;
				for(k=j+1; k<e_end && ev[k].timestamp<endtime; k++) {
					if(eop[k].thr!=eop[j].thr) continue;
					ops++;
					if(ops>warn_thundering_herd_max_intervening_ops)
						break;
					if(eop[k].cond==o_cond &&
					   (ev[k].event==ev_cond_wait ||
					    ev[k].event==ev_cond_timedwait))
					   	break;
				}
				if(ops<warn_thundering_herd_max_intervening_ops &&
				   k<e_end &&
				   ev[k].timestamp<endtime)
				{
					short_wakeups++;
					if(ev_thundering==0)
						ev_thundering = &ev[i];
				}
			}
		}
	}
	if(short_wakeups>0 &&
	   short_wakeups/(double)cond->accum_waiters_when_broadcast > warn_thundering_herd_ratio)
		suggest(ev_thundering,CONDTHUNDERINGHERD,
		        "(the reported event is just a sample)\n"
			"Condition %08lx were broadcasted %u times waking %lu threads, of which %lu went to sleep on the condition again shortly. This could indicate a \"thundering herd\" problem\n",
		        o_cond->addr, cond->accum_broadcast_operations, cond->accum_waiters_when_broadcast, short_wakeups);
}


static void analyze2(const EventRecord *ev, size_t events,
                     Object *o_cond, size_t conds,
		     Condition *cond,
		     EventObjectPointers *eop)
{
	for(size_t i=0; i<conds; i++) {
		//perf: signal without waiters
		if(cond[i].accum_signal_operations>0 &&
		   cond[i].accum_signal_ops_without_waiters>warn_signal_without_waiters_min_ops)
		{
			double sww_ratio = (double)cond[i].accum_signal_ops_without_waiters / cond[i].accum_signal_operations;
			if(sww_ratio>=warn_signal_without_waiters_ratio)
				warn(0,CONDSIGNALWITHOUTWAITERS,
				     "%5.0f%% of the pthread_cond_signal operations on condition %08lx were without waiting threads\n",
				     sww_ratio*100,
				     o_cond[i].addr
				    );
		}
		if(cond[i].accum_broadcast_ops_with_multiple_waiters>0 &&
		   cond[i].accum_waiters_when_broadcast>cond[i].accum_broadcast_operations)
			look_for_thundering_herd(ev,events,&o_cond[i],&cond[i],eop);
	}
//todo: perf: broadcast with many waiters which wait again very fast
}

static void dumpCondition(Object *o_cond, Condition *cond) {
	OutputParagraphStart();
	fprintf(fp_output,"Condition %08lx\n",o_cond->addr);
	fprintf(fp_output,"Created: ");
	if(o_cond->create_event)
		fprintf(fp_output,"%17.6f\n", o_cond->create_event->timestamp);
	else
		fprintf(fp_output,"(static condition)\n");
	fprintf(fp_output,"Destroyed: ");
	if(o_cond->destroy_event)
		fprintf(fp_output,"%17.6f\n", o_cond->destroy_event->timestamp);
	else
		if(o_cond->create_event)
			fprintf(fp_output,"never\n");
		else
			fprintf(fp_output,"(static mutex)\n");
	if(o_cond->name_event)
		fprintf(fp_output,"Name: %s\n",o_cond->name_event->name);
	fprintf(fp_output,"Wait+timedwaited operations: %u\n", cond->accum_wait_operations);
	fprintf(fp_output,"Timedwait operations: %u\n", cond->accum_timedwait_operations);
	fprintf(fp_output,"Failed timedwait operations: %u\n", cond->accum_failed_timedwait_operations);
	fprintf(fp_output,"signal operations: %u\n", cond->accum_signal_operations);
	fprintf(fp_output,"broadcast operations: %u\n", cond->accum_broadcast_operations);
	fprintf(fp_output,"signal operations without any waiters: %u\n", cond->accum_signal_ops_without_waiters);
	fprintf(fp_output,"broadcast operations with multiple waiters: %u\n", cond->accum_broadcast_ops_with_multiple_waiters);
	fprintf(fp_output,"accumulated waiters when broadcast: %lu\n", cond->accum_waiters_when_broadcast);
	OutputParagraphEnd();
}



struct HotCondition {
	unsigned accum_sb_operations;
	size_t i;
};

extern "C" {
static int cmp_hotcondition(const void *pv1, const void *pv2) {
	HotCondition *hc1=(HotCondition*)pv1;
	HotCondition *hc2=(HotCondition*)pv2;
	if(hc1->accum_sb_operations<hc2->accum_sb_operations)
		return -1;
	if(hc1->accum_sb_operations>hc2->accum_sb_operations)
		return  1;
	return 0;
}
}

static void dumpHotConditions(Object *o_cond, size_t conds,
                              Condition *cond
                             )
{
	HotCondition *hc = new HotCondition[conds];
	for(size_t i=0; i<conds; i++) {
		hc[i].accum_sb_operations = cond[i].accum_signal_operations+cond[i].accum_broadcast_operations;
		hc[i].i = i;
	}
	qsort(hc,conds,sizeof(*hc),cmp_hotcondition);
	
	fprintf(fp_output,"%ld most used conditions:\n",(long)dump_hot_condition_count);
	for(size_t i=conds<dump_hot_condition_count?0:conds-dump_hot_condition_count; i<conds; i++) {
		dumpCondition(o_cond+hc[i].i, cond+hc[i].i);
	}
	delete[] hc;
}

static void dump(Object *o_cond, size_t conds,
                 Condition *cond
                )
{
	for(size_t i=0; i<conds; i++) {
		dumpCondition(o_cond+i, cond+i);
	}
}


void analyze_condition_use(const EventRecord ev[], size_t events,
                           Object *o_cond, size_t conds,
		           EventObjectPointers *eop)
{
	OutputChapterStart("condition analysis");
	if(conds==0) {
		fprintf(fp_output,"(no conditions)\n");
		OutputChapterEnd("condition analysis");
		return;
	}
        
	Condition *cond = new Condition[conds];
	memset(cond,0,sizeof(Condition)*conds); //yeah, this is dirty
	//for(size_t i=0; i<conds; i++) {
	//	cond[i]. = 0.0;
	//}
	
	OutputSectionStart("condition use analysis");
	analyze1(ev,events,o_cond,conds,cond,eop);
	OutputSectionEnd("condition use analysis");
	
	OutputSectionStart("condition performance analysis");
	analyze2(ev,events,o_cond,conds,cond,eop);
	OutputSectionEnd("condition performance analysis");
	
	if(dump_hot_condition_count>0) {
		OutputSectionStart("dump of \"hot\" conditions");
		dumpHotConditions(o_cond,conds,cond);
		OutputSectionEnd("dump of \"hot\" conditions");
	}
	
	if(dump_conditions) {
		OutputSectionStart("dump of all conditions");
		dump(o_cond,conds,cond);
		OutputSectionEnd("dump of all conditions");
	}
	delete[] cond;
	
	OutputChapterEnd("condition analysis");
}
