//analyze_mutex
#include "die.h"
#include "analyze_mutex.h"
#include "ev.h"
#include <stdlib.h>
#include <string.h>
#include "analyze_output.h"
#include "mtanalyze.h"
#include "analyze_mutex_deadlocks.h"

struct Mutex {
	Object *owner_thr;
	unsigned lock_count;
	unsigned max_lock_count;
	const EventRecord *lock_event;
	unsigned long accum_lock_operations;
	unsigned long accum_contention_count;
	unsigned long accum_failed_lock_operations;
	unsigned waiters;
	unsigned max_waiters;
	const EventRecord *max_waiters_event;
	double accum_lock_wait_time;
	double max_lock_wait_time;
	const EventRecord *max_lock_wait_event;
	double accum_hold_time;
	double max_hold_time;
	const EventRecord *max_hold_time_event;
};


static bool improper_unlocking_found=false;

static void analyze1(const EventRecord ev[], size_t events,
                     Object *o_mtx, size_t mtxs,
		     Mutex *mtx,
		     EventObjectPointers *eop)
{
	for(size_t i=0; i<events; i++) {
		const EventRecord *ep = ev+i;
		Object *o=eop[i].mtx;
		if(!o) continue;
		Mutex *m = mtx+(o-o_mtx);
		switch(ep->event) {
			case ev_mutex_lock: {
				//increment accum_lock_operations
				//scan for <locked>
				//  update accum_lock_wait_time
				//  update max_lock_wait_time
				//if currently held, increment accum_contention_count
				//if recursive mutexes not allowed
				//  if owner=requestor ->bad
				//keep track of waiters
				m->accum_lock_operations++;
				
				size_t j;
				for(j=i; j!=events; j++) {
					if(eop[j].thr == eop[i].thr &&
					   ev[j].mtx_addr == ep->mtx_addr &&
					   ev[j].event==ev_mutex_locked)
						break;
				}
				if(j==events) {
					warn(ep,MUTEXLOCKNEVERRETURNED,
					     "pthread_mutex_lock() called but it never returned (cancelled thread? incomplete log?)\n");
					break;
				}
				
				double wt = ev[j].timestamp-ep->timestamp;
				m->accum_lock_wait_time += wt;
				if(wt>m->max_lock_wait_time) {
					m->max_lock_wait_time = wt;
					m->max_lock_wait_event = ep;
				}
				
				if(m->owner_thr)
					m->accum_contention_count++;
				
				if(!allow_recursive_mutexes) {
					if(m->owner_thr==eop[i].thr)
						die(ep,MUTEXRECURSIVE,
						    "pthread_mutex_lock() called for already held mutex (and recursive mutexes are not allowed)\n");
				}
				
				m->waiters++;
				if(m->waiters>m->max_waiters) {
					m->max_waiters = m->waiters;
					m->max_waiters_event = ep;
				}
				break;
			}
			case ev_mutex_trylock: {
				//increment accum_lock_operations
				//scan for <locked> or <trylock_failed>
				//  update accum_lock_wait_time
				//  update max_lock_wait_time
				//if currently held, increment accum_contention_count
				//if recursive mutexes not allowed
				//  if owner=requestor ->warn
				//keep track of waiters
				m->accum_lock_operations++;
				
				size_t j;
				for(j=i; j!=events; j++) {
					if(eop[j].thr == eop[i].thr &&
					   ev[j].mtx_addr == ep->mtx_addr &&
					   (ev[j].event==ev_mutex_locked || ev[j].event==ev_mutex_trylock_failed))
						break;
				}
				if(j==events) {
					warn(ep,MUTEXTRYLOCKNEVERRETURNED,
					     "pthread_mutex_trylock() called but it never returned (cancelled thread? incomplete log?)\n");
					break;
				}
				
				double wt = ev[j].timestamp-ep->timestamp;
				m->accum_lock_wait_time += wt;
				if(wt>m->max_lock_wait_time) {
					m->max_lock_wait_time = wt;
					m->max_lock_wait_event = ep;
				}
				
				if(m->owner_thr)
					m->accum_contention_count++;
				
				if(m->owner_thr==eop[i].thr) {
					if(m->lock_count>1)
						warn(ep,MUTEXTRYLOCKLOCKEDMUTEX,"pthread_mutex_trylock() called for already held recursive mutex (lock count = %u). This may not work as expected.\n",m->lock_count);
					else if(m->max_lock_count>1)
						warn(ep,MUTEXTRYLOCKLOCKEDMUTEX,"pthread_mutex_trylock() called for already held recursive mutex. This may not work as expected.\n",m->max_lock_count);
					else if(allow_recursive_mutexes)
						warn(ep,MUTEXTRYLOCKLOCKEDMUTEX,"pthread_mutex_trylock() called for already held mutex. This may not work as expected.\n");
					else
						warn(ep,MUTEXTRYLOCKLOCKEDMUTEX,"pthread_mutex_trylock() called for already held mutex.\n");
				}
				
				m->waiters++;
				if(m->waiters>m->max_waiters) {
					m->max_waiters = m->waiters;
					m->max_waiters_event = ep;
				}
				break;
			}
			case ev_mutex_locked: {
				//check info (rc)
				//update owner (+lock_count + max_lock_count)
				m->waiters--;
				if(ep->info!=0) {
					warn(ep,MUTEXINITFAILED,"pthread_mutex_lock() failed - this is unusual and probably indicates an error. Return value from pthread_mutex_lock() was %d\n",ep->info);
				} else {
					if(m->owner_thr==0) {
						m->owner_thr=eop[i].thr;
						m->lock_count=1;
						m->lock_event = ep;
						if(m->lock_count>m->max_lock_count)
							m->max_lock_count=m->lock_count;
					} else if(m->owner_thr==eop[i].thr) {
						m->lock_count++;
						if(m->lock_count>m->max_lock_count)
							m->max_lock_count=m->lock_count;
					} else {
						die(ep,MUTEXNOTMX,"pthread_mutex_lock/trylock() succeeded even though the lock is currently held by another thread (tid=%ld) (bug in pthread library?)\n",m->owner_thr->addr);
					}
				}
				break;
			}
			case ev_mutex_trylock_failed: {
				//update accum_failed_lock_operations
				m->accum_failed_lock_operations++;
				break;
			}
			case ev_mutex_unlock: {
				//check ownership
				//decrement lock_count (+owner_thr)
				//<scan backward for <locked>>
				//  update accum_hold_time
				if(eop[i].thr!=m->owner_thr) {
					if(m->owner_thr)
						die(ep,MUTEXUNLOCKOTHEROWNER,
						    "pthread_mutex_unlock() called for mutex which currently held by another thread (tid=%ld)\n",m->owner_thr->addr);
					else {
						improper_unlocking_found=true;
						warn(ep,MUTEXUNLOCKNOTOWNED,
						     "pthread_mutex_unlock() called for mutex which currently not held\n");
					}
				} else {
					m->lock_count--;
					if(m->lock_count==0) {
						m->owner_thr=0;
						double hold_time = ep->timestamp-m->lock_event->timestamp;
						m->accum_hold_time += hold_time;
						if(hold_time>m->max_hold_time) {
							m->max_hold_time = hold_time;
							m->max_hold_time_event = m->lock_event;
						}
					}
				}
				break;
			}
			case ev_mutex_unlocked: {
				//check info (rc)
				if(ep->info!=0)
					warn(ep,MUTEXUNLOCKERROR,
					     "pthread_mutex_unlock() returned an error. Very unusual."
					     "return code = %d\n", ep->info
					    );
				break;
			}
			case ev_cond_wait:
			case ev_cond_timedwait: {
				//check ownership
				//decrement lock_count (+owner_thr)
				//check lock_count
				//<scan backward for <locked>>
				//  update accum_hold_time
				if(eop[i].thr!=m->owner_thr) {
					if(m->owner_thr)
						die(ep,MUTEXCONDWAITOTHEROWNER,"pthread_cond_wait/timedwait() called with mutex owned by other thread (tid=%ld)\n",m->owner_thr->addr);
					else
						die(ep,MUTEXCONDWAITNOTOWNED,"pthread_cond_wait/timedwait() called with mutex not held\n");
				}
				if(m->lock_count!=1) {
					die(ep,MUTEXCONDWAITRECURSIVELOCKED,"pthread_cond_wait/timedwait() called with recursive mutex with lock level!=1 (%u)\n",m->lock_count);
				}
				if(implicit_cond_mutex_ops) {
					m->lock_count=0;
					m->owner_thr=0;
					double hold_time = ep->timestamp-m->lock_event->timestamp;
					m->accum_hold_time += hold_time;
					if(hold_time>m->max_hold_time) {
						m->max_hold_time = hold_time;
						m->max_hold_time_event = m->lock_event;
					}
				}
				break;
			}
			case ev_cond_waited:
			case ev_cond_timedwait_failed: {
				if(implicit_cond_mutex_ops) {
					//mutex owned?
					//  => bad/unexpected
					//re-lock the mutex
					if(m->owner_thr)
						die(ep,MUTEXCONDWAITEDOTHEROWNER,"pthread_cond_wait/timedwait() returned for mutex currently owned by other thread (tid=%ld)\n",m->owner_thr->addr);
					m->owner_thr = eop[i].thr;
					m->lock_count = 1;
					m->accum_lock_operations++; //yes, this should be counted too
					m->lock_event = ep;
				}
				break;
			}
			case ev_mutex_name:
			case ev_mutex_create:
			case ev_mutex_created:
			case ev_mutex_destroy:
			case ev_mutex_destroyed:
				break;
			default:
				die(ep,MUTEX_UNKNOWN_OP,"(internal error) unhandled event in mutex analyze\n");
		}
	}
	//check for still held mutexes
	for(size_t i=0; i<mtxs; i++) {
		if(mtx[i].lock_count>0) {
			warn(0,MUTEXSTILLLOCKED,
			       "Mutex still locked at end of event log, mutex=%08lx, thread=%ld\n"
			       "(Proper cleanup will make this warning go away)\n",
			       mtx[i].lock_event->mtx_addr, mtx[i].owner_thr->addr);
		}
	}
}


static void analyze2(const EventRecord ev[], size_t events,
                     Object *o_mtx, size_t mtxs,
		     Mutex *mtx,
		     EventObjectPointers *eop)
{
	for(size_t i=0; i<mtxs; i++) {
		if(mtx[i].accum_lock_operations==0) continue;
		
		//non-trivial contention (contention/ops > x%)
		double contention_ratio = mtx[i].accum_contention_count / (double)mtx[i].accum_lock_operations;
		if(contention_ratio>=warn_mutex_contention_ratio)
			warn(0, MUTEXHIGHCONTENTION,
			     "High contention ratio for mutex %08lx (%5.2f%% lock operations were contended)\n",
			     o_mtx[i].addr,
			     contention_ratio*100
			    );
		
		//long lock time
		if(mtx[i].max_lock_wait_time>warn_mutex_long_lock_time)
			warn(mtx[i].max_lock_wait_event, MUTEXLONGLOCKTIME,
			     "In at least one occurrence it took %.3f seconds to lock mutex\n",
			     mtx[i].max_lock_wait_time
			    );
		
		//long hold time
		if(mtx[i].max_hold_time>warn_mutex_long_hold_time)
			warn(mtx[i].max_hold_time_event, MUTEXLONGHOLDTIME,
			     "Mutex %08lx were held for %.3f seconds here\n",
			     o_mtx[i].addr,mtx[i].max_hold_time
			    );
		
		//longer-than-usual lock time
		double avg_lock_time = mtx[i].accum_lock_wait_time/mtx[i].accum_lock_operations;
		if(mtx[i].accum_lock_operations>=3 &&
		   mtx[i].max_lock_wait_time>avg_lock_time*warn_mutex_longer_lock_time)
			warn(mtx[i].max_lock_wait_event,MUTEXLONGERTHANUSUALLOCKTIME,
			     "On at least one occurrence it took %.3f seconds to lock the mutex (average=%.3f seconds)\n",
			     mtx[i].max_lock_wait_time,
			     avg_lock_time
			    );
		
		//many failed trylocks
		double failed_trylock_ratio = (double)mtx[i].accum_failed_lock_operations / mtx[i].accum_lock_operations;
		if(failed_trylock_ratio>warn_failed_trylock_ratio)
			warn(0,MUTEXMANYFAILEDTRYLOCK,
			     "Mutex %08lx has many failed pthread_mutex_trylock() operations (%ld failed out of %ld operations)\n",
			     o_mtx[i].addr,
			     mtx[i].accum_failed_lock_operations,
			     mtx[i].accum_lock_operations
			    );
		
		//deep locks
		if(mtx[i].max_lock_count>=warn_deep_locks)
			suggest(0,MUTEXDEEPRECURSIVELOCK,
			        "Recursive mutex %08lx has deep lock counts (max=%ld)\n"
			        "Consider using a non-recusive mutex and locking at start of operations\n",
			        o_mtx[i].addr,
			        mtx[i].max_lock_count
			       );
		
		//locks used by only one thread
		if(mtx[i].accum_lock_operations >= analyze_singlethread_mutex_threshold) {
			bool used_by_one_thread=true;
			Object *thr=0;
			size_t start = o_mtx[i].create_event?(ev-o_mtx[i].create_event):0;
			size_t end = o_mtx[i].destroy_event?(ev-o_mtx[i].destroy_event)+1:events;
			for(size_t j=start; j<end; j++) {
				if(eop[j].mtx!=&o_mtx[i]) continue;
				if(eop[j].thr!=thr) {
					if(thr==0)
						thr=eop[j].thr;
					else {
						used_by_one_thread=false;
						break;
					}
				}
			}
			if(used_by_one_thread) {
				if(thr)
					suggest(0,MUTEXONETHREAD,
					        "Mutex %08lx is only used by thread %ld. It may be possible to eliminate the mutex\n",
					        o_mtx[i].addr,
					        thr->addr
					       );
				else
					suggest(0,MUTEXNEVERUSED,
					        "Mutex %08lx is never used. It is possible to eliminate the mutex\n",
					        o_mtx[i].addr
					       );
			}
		}
		
		//many waiters
		if(mtx[i].max_waiters>warn_mutex_many_waiters_threshold)
			warn(mtx[i].max_waiters_event,MUTEXMANYWAITERS,
			     "The mutex has many waiting threads. At this event there is %u waiting threads\n",
			     mtx[i].max_waiters
			    );
	}
}


static void dumpMutex(Object *o_mtx, Mutex *mtx) {
	OutputParagraphStart();
	fprintf(fp_output,"Mutex %08lx\n",o_mtx->addr);
	fprintf(fp_output,"Created: ");
	if(o_mtx->create_event)
		fprintf(fp_output,"%17.6f\n", o_mtx->create_event->timestamp);
	else
		fprintf(fp_output,"(static mutex)\n");
	fprintf(fp_output,"Destroyed: ");
	if(o_mtx->destroy_event)
		fprintf(fp_output,"%17.6f\n", o_mtx->destroy_event->timestamp);
	else
		if(o_mtx->create_event)
			fprintf(fp_output,"never\n");
		else
			fprintf(fp_output,"(static mutex)\n");
	if(o_mtx->name_event)
		fprintf(fp_output,"Name: %s\n",o_mtx->name_event->name);
	fprintf(fp_output,"Max lock depth: %d\n", mtx->max_lock_count);
	fprintf(fp_output,"Total lock/trylock operations: %lu\n", mtx->accum_lock_operations);
	fprintf(fp_output,"Total contented operations: %lu\n", mtx->accum_contention_count);
	fprintf(fp_output,"Failed pthread_mutex_trylock() calls: %lu\n", mtx->accum_failed_lock_operations);
	fprintf(fp_output,"Max. threads wanting mutex: %u\n",mtx->max_waiters);
	fprintf(fp_output,"Total lock->locked time: %5.3f\n", mtx->accum_lock_wait_time);
	fprintf(fp_output,"Max. lock->locked time: %5.3f\n", mtx->max_lock_wait_time);
	fprintf(fp_output,"Accumulated hold time: %5.3f\n", mtx->accum_hold_time);
	fprintf(fp_output,"Max hold time: %5.3f\n", mtx->max_hold_time);
	OutputParagraphEnd();
}

struct HotMutex {
	unsigned long accum_lock_operations;
	size_t i;
};

extern "C" {
static int cmp_hotmutex(const void *pv1, const void *pv2) {
	HotMutex *hm1=(HotMutex*)pv1;
	HotMutex *hm2=(HotMutex*)pv2;
	if(hm1->accum_lock_operations<hm2->accum_lock_operations)
		return -1;
	if(hm1->accum_lock_operations>hm2->accum_lock_operations)
		return  1;
	return 0;
}
}


static void dumpHotMutexes(Object *o_mtx, size_t mtxs,
                           Mutex *mtx
                          )
{
	HotMutex *hm = new HotMutex[mtxs];
	for(size_t i=0; i<mtxs; i++) {
		hm[i].accum_lock_operations = mtx[i].accum_lock_operations;
		hm[i].i = i;
	}
	qsort(hm,mtxs,sizeof(*hm),cmp_hotmutex);
	
	fprintf(fp_output,"%ld most used mutexes:\n",(long)dump_hot_mutex_count);
	for(size_t i=mtxs<dump_hot_mutex_count?0:mtxs-dump_hot_mutex_count; i<mtxs; i++) {
		fprintf(fp_output,"---\n");
		dumpMutex(o_mtx+hm[i].i, mtx+hm[i].i);
	}
	delete[] hm;
}

static void dump(Object *o_mtx, size_t mtxs,
                 Mutex *mtx
                )
{
	for(size_t i=0; i<mtxs; i++) {
		dumpMutex(o_mtx+i, mtx+i);
	}
}


void analyze_mutex_use(const EventRecord ev[], size_t events,
                       Object *o_mtx, size_t mtxs,
		       EventObjectPointers *eop)
{
	OutputChapterStart("mutex analysis");
	
	Mutex *mtx = new Mutex[mtxs];
	memset(mtx,0,sizeof(Mutex)*mtxs);
	for(size_t i=0; i<mtxs; i++) {
		mtx[i].accum_lock_wait_time=0.0;
		mtx[i].max_lock_wait_time=0.0;
		mtx[i].accum_hold_time=0.0;
		mtx[i].max_hold_time=0.0;
	}
	
	OutputSectionStart("mutex use analysis");
	analyze1(ev,events,o_mtx,mtxs,mtx,eop);
	OutputSectionEnd("mutex use analysis");
	
	OutputSectionStart("mutex performance analysis");
	analyze2(ev,events,o_mtx,mtxs,mtx,eop);
	OutputSectionEnd("mutex performance analysis");
	
	if(perform_deadlock_analysis) {
		if(!improper_unlocking_found) {
			OutputSectionStart("mutex deadlock analysis");
			analyze_mutex_deadlocks(ev,events,o_mtx,mtxs,eop);
			OutputSectionEnd("mutex deadlock analysis");
		} else
			fprintf(fp_output,"Deadlock analysis skipped due to improper unlocking\n");
	}
	
	if(dump_hot_mutex_count>0) {
		OutputSectionStart("dump of \"hot\" mutexes");
		dumpHotMutexes(o_mtx,mtxs,mtx);
		OutputSectionEnd("dump of \"hot\" mutexes");
	}
	
	if(dump_mutexes) {
		OutputSectionStart("dump of all mutexes");
		dump(o_mtx,mtxs,mtx);
		OutputSectionEnd("dump of all mutexes");
	}
	delete[] mtx;
	
	OutputChapterEnd("mutex analysis");
}
