//analyze_condition
#include "die.h"
#include "analyze_thread.h"
#include "ev.h"
#include <stdlib.h>
#include <string.h>
#include "analyze_output.h"
#include "mtanalyze.h"
#include "level_history.h"
#include <limits.h>

#ifndef _POSIX_THREAD_THREADS_MAX
//HP/UX is missing this one
#define _POSIX_THREAD_THREADS_MAX 64
#endif

struct Thread {
	unsigned long thread_create_operations;
	unsigned long failed_thread_create_operations;
	unsigned long cond_signal_operations;
	unsigned long cond_broadcast_operations;
	unsigned long cond_wait_operations;
	unsigned long mutex_operations;
};


static void analyze1(const EventRecord ev[], size_t events,
                     Object *o_thr, size_t /*thrs*/,
                     Thread *thr,
                     EventObjectPointers *eop)
{
	for(size_t i=0; i<events; i++) {
		const EventRecord *ep = ev+i;
		Object *o=eop[i].thr;
		//no check for pointer as it is always set
		Thread *t = thr+(o-o_thr);
		switch(ep->event) {
			case ev_thread_create: {
				t->thread_create_operations++;
				break;
			}
			case ev_thread_create_failed: {
				t->failed_thread_create_operations++;
				break;
			}
			case ev_cond_signal: {
				t->cond_signal_operations++;
				break;
			}
			case ev_cond_broadcast: {
				t->cond_broadcast_operations++;
				break;
			}
			case ev_cond_wait:
			case ev_cond_timedwait: {
				t->cond_wait_operations++;
				break;
			}
			case ev_mutex_lock:
			case ev_mutex_trylock:
			case ev_mutex_unlock: {
				t->mutex_operations++;
				break;
			}
			default:
				;
		} //switch
	}
}

static void analyze2(const EventRecord ev[], size_t events,
                     Object *o_thr, size_t thrs,
		     Thread *,
		     EventObjectPointers *)
{
	{
		//perf: high peak
		//port: peak>minimum maxima
		LevelHistory lh;
		for(size_t i=0; i<thrs; i++) {
			unsigned from_idx;
			if(o_thr[i].create_event)
				from_idx = (unsigned)(o_thr[i].create_event-ev);
			else
				from_idx = 0;
			unsigned to_idx;
			if(o_thr[i].destroy_event)
				to_idx = (unsigned)(o_thr[i].destroy_event-ev);
			else
				to_idx = (unsigned)events;
			lh.increment(from_idx,to_idx);
		}
		unsigned max_threads = 0;
		unsigned max_threads_ev_idx = 0;
		for(size_t i=0; i<lh.edges(); i++) {
			unsigned ev_idx = lh.edgeTime(i);
			unsigned level = lh.edgeLevel(i);
			if(level>max_threads) {
				max_threads = level;
				max_threads_ev_idx = ev_idx;
			}
		}
		if(max_threads>warn_concurrent_threads_threshold)
			warn(&ev[max_threads_ev_idx], MANYCONCURRENTTHREADS,
			     "At this point %u threads were running. This is a lot.\n",
			     max_threads
			    );
		if(max_threads>_POSIX_THREAD_THREADS_MAX)
			suggest(&ev[max_threads_ev_idx], NONPORTABLECONCURRENTTHREADS,
			        "At this point %u threads were running, which exceeds _POSIX_THREAD_THREADS_MAX. This is not portable\n",
			        max_threads
			       );
	}
	
	//perf: lots of shortlived threads
	double program_lifetime = ev[events-1].timestamp - ev[0].timestamp;
	unsigned shortlived_threads=0;
	for(size_t i=0; i<thrs; i++) {
		double start,end;
		start = o_thr[i].create_event?o_thr[i].create_event->timestamp:ev[0].timestamp;
		end = o_thr[i].destroy_event?o_thr[i].destroy_event->timestamp:ev[events-1].timestamp;
		double lifetime = end-start;
		if(lifetime/program_lifetime < warn_shortlived_thread_ratio) {
			suggest(o_thr[i].create_event,
			        SHORTLIVEDTHREAD,
			        "This thread were short-lived (%.6f seconds)\n",
			        lifetime
			       );
		}
	}
	if(shortlived_threads >= thrs*warn_shortlived_threads_ratio)
		warn(0, MANYSHORTLIVEDTHREAD,
		     "%u short-lived threads were found. This could indicate 'thread-per-connection' implementation which in general does not scale as well as 'worker pool'\n",
		     shortlived_threads
		    );
}

static void dumpThread(Object *o_thr, Thread *thr) {
	OutputParagraphStart();
	fprintf(fp_output,"Thread %ld\n",o_thr->addr);
	fprintf(fp_output,"Created: ");
	if(o_thr->create_event)
		fprintf(fp_output,"%17.6f\n", o_thr->create_event->timestamp);
	else
		fprintf(fp_output,"(at program startup)\n");
	fprintf(fp_output,"Destroyed: ");
	if(o_thr->destroy_event)
		fprintf(fp_output,"%17.6f\n", o_thr->destroy_event->timestamp);
	else
		fprintf(fp_output,"never\n");
	if(o_thr->create_event && o_thr->destroy_event)
		fprintf(fp_output,"Lifetime: %9.6f seconds\n", o_thr->destroy_event->timestamp-o_thr->create_event->timestamp);
	if(o_thr->name_event)
		fprintf(fp_output,"Name: %s\n",o_thr->name_event->name);
	fprintf(fp_output,"pthread_create() by this thread: %lu\n", thr->thread_create_operations);
	fprintf(fp_output,"Failed pthread_create() by this thread: %lu\n", thr->failed_thread_create_operations);
	fprintf(fp_output,"pthread_cond_signal operations: %lu\n", thr->cond_signal_operations);
	fprintf(fp_output,"pthread_cond_broadcast operations: %lu\n", thr->cond_broadcast_operations);
	fprintf(fp_output,"pthread_cond_wait/timedwait operations: %lu\n", thr->cond_wait_operations);
	fprintf(fp_output,"Mutex operations (lock+unlock): %lu\n", thr->mutex_operations);
	OutputParagraphEnd();
}


static void dump(Object *o_thr, size_t thrs,
                 Thread *thr
                )
{
	for(size_t i=0; i<thrs; i++) {
		dumpThread(o_thr+i, thr+i);
	}
}


void analyze_thread_use(const EventRecord ev[], size_t events,
                        Object *o_thr, size_t thrs,
                        EventObjectPointers *eop)
{
	OutputChapterStart("thread analysis");
	Thread *thr = new Thread[thrs];
	memset(thr,0,sizeof(Thread)*thrs); //yeah, this is dirty
	//for(size_t i=0; i<thrs; i++) {
	//	thrs[i]. = 0.0;
	//}
	
	OutputSectionStart("thread use analysis");
	analyze1(ev,events,o_thr,thrs,thr,eop);
	OutputSectionEnd("thread use analysis");
	
	OutputSectionStart("thread performance analysis");
	analyze2(ev,events,o_thr,thrs,thr,eop);
	OutputSectionEnd("thread performance analysis");
	
	if(dump_threads) {
		OutputSectionStart("dump of all threads");
		dump(o_thr,thrs,thr);
		OutputSectionEnd("dump of all threads");
	}
	delete[] thr;
	OutputChapterEnd("thread analysis");
}
